from __future__ import annotations

import abc
from enum import Enum
import functools
import itertools
import json
import os
from pathlib import Path
import re
import subprocess
import sys
from typing import Collection, Dict, List, Optional, Union
from urllib.parse import urlparse
import warnings

from packaging.specifiers import SpecifierSet
from pipreqs import pipreqs
import pkg_resources
from pkg_resources import Requirement
import yaml

from grid.sdk import env
from grid.sdk.source_code.git import execute_git_command

PIP_MANAGER = 'PIP'
CONDA_MANAGER = 'CONDA'


def find_package_manager(file):
    """
    Find the package manager for the provided dependency file depends on basic heuristics.
    We currently assume a minimal conda environment file would have `dependencies` as the
    key that lists all the dependencies, as described here
    https://conda.io/projects/conda/en/latest/user-guide/tasks/manage-environments.html#create-env-file-manually
    """
    if not file:
        raise RuntimeError("Invalid call to find the package manager")
    f = Path(file).open()
    try:
        content = yaml.safe_load(f)
    except Exception:
        # Assuming PIP if yaml load fails
        return PIP_MANAGER
    else:
        if isinstance(content, dict):
            # valid conda env file should have `dependencies` keys
            if 'dependencies' in content:
                return CONDA_MANAGER
    return PIP_MANAGER


PIP_REQUIREMENTS_FILE = "requirements.txt"
CONDA_REQUIREMENTS_FILE = "environment.yml"


def get_manager(repo_root) -> DependencyManagerBase:
    files = os.listdir(repo_root)
    conda_env = os.getenv("CONDA_DEFAULT_ENV")
    if CONDA_REQUIREMENTS_FILE in files:
        if PIP_REQUIREMENTS_FILE in files:  # skipcq: PYL-R1705
            warnings.warn(
                f"Found both {CONDA_REQUIREMENTS_FILE} and {PIP_REQUIREMENTS_FILE}. "
                f"Defaulting to use {PIP_REQUIREMENTS_FILE}."
            )
            manager_class = PipManager
        else:
            manager_class = CondaManager
    elif PIP_REQUIREMENTS_FILE in files:
        manager_class = PipManager
    elif conda_env:
        manager_class = CondaManager
    else:
        manager_class = PipManager
    return manager_class()


def guess_manager_from_dependency_file(dependency_file) -> DependencyManagerBase:
    """
    If user provides the dependency file instead of letting us find it from the root,
    use this function to create the Manager. If dependency_file is not provided,
    use `_get_manager`
    """
    package_manager = find_package_manager(dependency_file)
    if package_manager == PIP_MANAGER:
        return PipManager(dependency_file=dependency_file)
    if package_manager == CONDA_MANAGER:
        return CondaManager(dependency_file=dependency_file)

    raise RuntimeError("Unknown package manager")


# def serialize_dependencies(config_path: Optional[str] = None, dependency_file: Optional[str] = None) -> Optional[Path]:
#     """
#     Serialize dependency information to the corresponding source.
#
#     Write python dependencies to the corresponding dependency listing
#     (requirements.txt or environment.yml) and system dependencies to
#     grid config (config.yml). Both write operations are actually offloaded
#     downstream to the dependency manager object. Also fetches warnings
#     from the dependency manager and echo to the user's terminal
#
#     config:
#         Grid config either read from the file provided by the user or the
#         default generated one
#
#     Returns
#     -------
#     Path to the requirement file
#     """
#     check_status = check_dependency_listing(do_diff_check=False, ignore_warnings=False)
#     if not check_status:
#         print("Dependency check failed. Cannot sync environment")
#         return None
#
#     repo_root = execute_git_command(['rev-parse', '--show-toplevel'])
#     try:
#         if dependency_file:
#             deps_manager = guess_manager_from_dependency_file(dependency_file)
#         else:
#             deps_manager = get_manager(repo_root)
#     except Exception as e:  # noqa
#         # it's already being checked by `check_dependency_listing`
#         raise RuntimeError("Exception while initializing dependency manager")
#
#     deps_manager.write_spec()
#     deps_manager.write_config(config_path)
#     return deps_manager.req_file


def check_dependency_listing(
    repo_root: Path,
    do_diff_check=True,
    ignore_warnings=env.IGNORE_WARNINGS,
    dependency_file=None,
):
    """
    Evaluate if dependency listing is correct or not

    Determines the package manager employed by the user to specify
    application dependencies, setting up the appropriate interaction
    classes used to query one of the various package managers available
    as the system resolves dependency versions. Should no requirement file
    exist which specifies needed libraries+versions, this method provides
    the logic through which users are asked to select how they would like
    to proceed - automatic heuristic based generation or manual.

    Parameters
    ----------
    do_diff_check: bool
        Check for the diff between requirements file and the environment. This
        is True by default but `serialize_dependencies` checks this explicitly
        as part of serializing the dependencies and hence will set as False
    ignore_warnings: bool
        Should warnings be ignored. Should be removed when we move
        click component out of SDK -> TODO
    repo_root
        Root of the repository. Defaults to git root

    Returns
    -------
    Return False if check fails. True otherwise
    """
    repo_root = Path(execute_git_command(['rev-parse', '--show-toplevel'])).absolute()
    if ignore_warnings:
        return True

    # Assuming the requirements file is always at the repo root if not passed
    # through dependency_file argument
    files = os.listdir(repo_root)

    try:
        if dependency_file:
            deps_manager = guess_manager_from_dependency_file(dependency_file)
        else:
            deps_manager = get_manager(repo_root)
    except Exception as e:  # noqa
        # ==============================================
        # Case 1: Initializing dependency manager failed
        # ==============================================
        # Ignoring to avoid internal errors to go to users
        if PIP_REQUIREMENTS_FILE not in files and CONDA_REQUIREMENTS_FILE not in files:
            message = "No requirements.txt or environment.yml found. Your build could crash or not start."
            warnings.warn(message)
        return False

    # =====================================
    # Case 2: Requirements file not present
    # =====================================
    if PIP_REQUIREMENTS_FILE not in files and CONDA_REQUIREMENTS_FILE not in files:
        deps_found = "\n        ".join(deps_manager.source_deps)
        message = (
            "No requirements.txt or environment.yml found but we identified below "
            "dependencies from your source. Your build could crash or not start. "
            f"\n {deps_found} \n"
        )
        warnings.warn(message)
        return False

    # ===========================================================
    # Case 3: Packages found imported for which spec is not found
    # ===========================================================
    missing = "\n        ".join(deps_manager.get_missing())
    if missing:
        message = (
            "Your requirements.txt or environment.yml are missing the following packages:"
            f"\n {missing} \n"
            f"Your experiment could crash if these are missing! Add them yourself or try `grid sync-env`"
        )
        warnings.warn(message)
        return False

    # ==============================================
    # Case 4: Incomplete/different-spec requirements
    # ==============================================
    if do_diff_check and deps_manager.has_change:
        message = (
            "Incomplete requirements.txt or environment.yml found. "
            "Your build could crash or not start. Run `grid sync-env` "
            "to auto-populate the changed packages"
        )
        warnings.warn(message)
        return False

    # Returning True if all looks good
    return True


# https://www.python.org/dev/peps/pep-0610/
PACKAGE_URL_METAFILE = "direct_url.json"
ALLOWED_SCHEMES = {'http', 'https', 'file', 'ftp', 'ssh', 'git', 'sftp', 'git+http', 'git+https'}


class NonComparableError(Exception):
    """
    Two Packages are non-comparable
    """


def execute_conda_command(args: List[str]):
    process = subprocess.run(["conda"] + args,
                             stdout=subprocess.PIPE,
                             stderr=subprocess.PIPE,
                             universal_newlines=True,
                             check=False)
    output = process.stdout.strip() + process.stderr.strip()
    return output


def get_url_scheme(url: str) -> str:
    """Parse the URL and get the scheme"""
    return urlparse(url).scheme


def is_valid_url(url: str) -> bool:
    """Check the validity of the URL. Currently it's only based
    on the scheme. Whether the scheme is in allowed schemes or not
    """
    return get_url_scheme(url) in ALLOWED_SCHEMES


class RuleLabels(Enum):
    PIP_NAME_MAPPING = "pip_name_mapping"
    CONDA_NAME_MAPPING = "conda_name_mapping"
    SYSTEM = "system"


@functools.lru_cache(maxsize=None)
def _get_rule_maps() -> Dict[str, Dict[str, str]]:
    """
    Parsing the rule files for mapping python package names, conda names
    and finding out the system dependencies. The whole rule set is built by
    keeping the parsed pypi name as reference. The parsed pypi name is the
    name of a package as listed in pypi but replaced all the hyphens with
    underscores. Below given are few examples of parsed pypi name

            Import name : sklearn
            Parsed pypi : scikit_learn
            pypi name   : scikit-learn

    For more details, refer the documentation in the YAML rule file
    """
    rule_file = Path(__file__).parent / "dependency_rules/dependency_rules.yml"
    mappings = yaml.safe_load(rule_file.open())
    mandatory_keys = {val.value for val in RuleLabels}
    # Both must have exactly the same elements
    if mandatory_keys.symmetric_difference(mappings.keys()):
        raise RuntimeError("Rule file is corrupted")
    out = {}
    for k in mandatory_keys:
        out[k] = dict(mapping.strip().split(":") for mapping in mappings[k])

    # nested parsing of system dependencies
    for name, value in out[RuleLabels.SYSTEM.value].items():
        out[RuleLabels.SYSTEM.value][name] = value.split(",")

    return out


def _get_system_deps(name: str) -> List[str]:
    """
    Get the list of system dependencies required for a python package. For
    example graphviz is required to have graphviz and xdg-utils installed
    in the system apart from the python package itself.

    Parameters
    ----------
    name:
        Parsed pypi name

    Returns
    -------
    List of system dependencies required for a python package to be working
    """
    system_rules = _get_rule_maps()[RuleLabels.SYSTEM.value]
    return system_rules.get(name, [])


def _conda2parsed_pypi(name: str) -> str:
    """
    Parsed pypi names are being used as the standard for making comparisons
    in this module. For instance, conda list the package pytorch as
    "pytorch" while pip list it as "torch". This function converts conda name
    ("pytorch" to parsed pypi name ("torch") using the predefined mapping rule
    """
    conda_rules = _get_rule_maps()[RuleLabels.CONDA_NAME_MAPPING.value]
    return conda_rules.get(name, name).replace("-", "_")


def _source2parsed_pypi(name: str) -> str:
    """
    Parsed pypi names are being used as the standard for making comparisons
    in this module. This function converts the `import` name to the parsed
    pypi name. While both names are same for almost all the cases, there are
    few that isn't. For example sklearn is the import name while scikit-learn
    is the pypi name (and scikit_learn is the parsed pypi name - note the
    underscore). Note that for python's dependency manager, both "-" and "_"
    has same meaning. scikit-learn = scikit_learn
    """
    pypi_rules = _get_rule_maps()[RuleLabels.PIP_NAME_MAPPING.value]
    return pypi_rules.get(name, name).replace("-", "_")


def _get_distribution_url(dist: pkg_resources.Distribution) -> Optional[str]:
    """
    Fetch the distribution URL from package distribution based on pep440
    and pep610. Different details about how this feature eventually landed
    in `pip` and different considerations can be found at the issue page -
    https://github.com/pypa/pip/issues/609.
    """
    try:
        url_meta = json.loads(dist.get_metadata(PACKAGE_URL_METAFILE))
    except (FileNotFoundError, ValueError, UnicodeDecodeError, KeyError):
        return None
    url = url_meta.get("url")
    if not url:
        return None
    if "vcs_info" in url_meta:
        commit_id = url_meta["vcs_info"].get("commit_id")
        vcs = url_meta["vcs_info"].get("vcs")
        return f"{vcs}+{url}@{commit_id}"
    return url


def _is_dist_editable(dist: pkg_resources.Distribution) -> bool:
    """
    Check if a Distribution is installed as editable
    """
    for path_item in sys.path:
        egg_link = Path(path_item) / dist.project_name / '.egg-link'
        if egg_link.is_file():
            return True
    return False


def _get_url_n_rev(string):
    """ If the url is in the format `url@head

    Example: https://github.com/org/repo@branch
    """
    if "@" in string:
        return string.rsplit("@", maxsplit=1)
    return string, None


def _get_valid_stem(string: str) -> Optional[str]:
    """
    Validate and get the stem from the given path / URL

    It verifies whether the given path is valid or not and if valid
    fetches the stem from the path and consider that as the name. This
    doesn't work in all the cases (eg. hangar-py) but it is probably the
    most reasonable assumption that would work in almost all the cases
    """
    string = string.strip()
    scheme = get_url_scheme(string).lower()
    if scheme == "file":
        raise RuntimeError(f"Local paths are not portable '{string}'")

    # Windows paths either get parsed as scheme = drive letter (e.g. "C") for absolute paths or as scheme = string for
    # relative paths. We explicitly filter these cases before checking for a match to ALLOWED_SCHEMES.
    if scheme and len(scheme) > 1 and scheme != string and scheme not in ALLOWED_SCHEMES:
        raise RuntimeError(f"Valid URL schemes are {ALLOWED_SCHEMES}")

    if is_valid_url(string):
        url, _ = _get_url_n_rev(string)
        parsed = urlparse(url)
        if parsed.path:
            path = Path(parsed.path)
        else:
            raise RuntimeError(f"Invalid URL '{string}'")

    else:
        path = Path(string)
        if path.is_absolute():
            raise RuntimeError(f"Absolute paths might not resolve '{path}'")
        if path.exists():
            cwd = Path.cwd()
            # Use os.path.abspath as an alternative to path.resolve() which doesn't work on Windows.
            absolute = Path(os.path.abspath(path))
            if cwd not in absolute.parents and cwd != absolute:
                raise RuntimeError(f"Paths outside of the current working directory are not portable: {path}")
        else:
            raise RuntimeError(f"Error while parsing {string}")
    return path.resolve().stem.split(".")[0]


def _guess_valid_name(string):
    """
    Heuristics to fallback when `Requirement.parse` fails. This function tries
    to find a valid name from the given string and return. This name is
    then used to make comparison from the names from other sources.

    Evidently, pip's parsing logic is exhaustive and we wouldn't want to copy
    the whole pip parser to our source code. The heuristics this function
    supports are a given below. Every other string would lead to a
    warning or an exception. More details on parsing logic, checkout pip's
    source code or the grammar + parser specified in pip508

        location: /src/pip/_internal/req/constructors.py#L291
        commit: a089883a8

        - Current directory
        - Wheel files or sub directories in current directory
        - Valid URLs (git, remote wheel file, remote zip file etc)
        - Valid Requirement spec with options (pkg==0.1 -i <url>)
    """
    string = string.strip()
    if string.startswith("-e "):
        raise RuntimeError("Found editable requirements which might not be" f"portable {string}")
    if "@ " in string:
        # name@ path is not a valid specification in 3.7
        name, _ = string.split("@ ", maxsplit=1)
        if not name:
            raise RuntimeError(f"Error while parsing '{string}'")
        return name.strip()
    marker_sep = '; ' if is_valid_url(string) else ";"
    substr = string.split(marker_sep, 1)[0]
    # If options are present in the string (pkg=0.1 -i <url>)
    splits = re.split(r" -[\w\-]+ ", substr)
    if len(splits) > 1:
        try:
            req = Requirement.parse(splits[0])
        except Exception:  # noqa
            raise RuntimeError(f"Error while parsing {string}")
        name = req.project_name
    else:
        name = _get_valid_stem(substr)
    return name


class BasePackage(abc.ABC):
    """Base Package class for holding the PIP and Conda Spec.

    Custom Package classes must be inherited from this class to include
    custom logic for equality check and parsing

    Note: Parameter `name` comes from the name of the package defined
    by the developer of the package. But in cases where we cannot infer
    that name, we fallback to other methods. For example, in the case
    of a URL, the name is the `stem` of the `path` in the URL. If the
    name is identified by none of the methods, we generate a uuid instead
    """
    def __init__(self, name: str):
        self.name = name
        self.parsed_name = self.name.replace("-", "_")

    @abc.abstractmethod
    def from_string(self, string):
        raise NotImplementedError

    @abc.abstractmethod
    def __eq__(self, other: BasePackage) -> bool:
        return NotImplemented

    @abc.abstractmethod
    def __str__(self):
        raise NotImplementedError


class CondaPackage(BasePackage):
    """
    Do not call the constructor / init directly. Use classmethod(s) instead
    """
    def __init__(self, name: Optional[str], version: Optional[str], build: Optional[str]):
        super().__init__(name)
        self.parsed_name = _conda2parsed_pypi(self.name)
        self.version = version
        self.build = build

    @classmethod
    def from_string(cls, string: str) -> CondaPackage:
        splits = re.split(r"\W*=\W*", string)
        if len(splits) > 3:
            raise RuntimeError(f"ParsingError: Conda spec {string} is " f"not readable")
        expected_elem_count = 3  # name, version and build
        splits = splits + [None] * (expected_elem_count - len(splits))  # noqa
        name, version, build = splits
        return cls(name=name, version=version, build=build)

    def __str__(self):
        not_none = [val for val in (self.name, self.version, self.build) if val]
        return "=".join(not_none)

    def __eq__(self, other: CondaPackage):
        if other.__class__ is self.__class__:
            return all((self.version == other.version, self.parsed_name == other.parsed_name))
        return NotImplemented


class PipPackage(BasePackage):
    """
    Offload all parsing, de-parsing, checks etc to `Requirement` object.
    Do not call the constructor / init directly. Use classmethod(s) instead
    """
    def __init__(self, name: str, req_obj: Optional[Requirement] = None, original_string: Optional[str] = None):
        super().__init__(name)
        self._requirement_obj = req_obj
        self._original_string = original_string

    @classmethod
    def from_string(cls, string) -> PipPackage:
        string = string.strip()
        try:
            req = Requirement.parse(string)
        except Exception:  # noqa
            # generic exception since different python versions have
            # inconsistencies in raising exceptions
            obj = None
        else:
            obj = cls.from_requirement(req)

        if obj is None:
            name = _guess_valid_name(string)
            if name is None:
                raise RuntimeError(f"Guessing name for '{string}' failed")
            obj = cls(name=name, original_string=string)
        return obj

    @classmethod
    def from_requirement(cls, req: pkg_resources.Requirement):
        self = cls(req.project_name, req_obj=req)
        return self

    @classmethod
    def from_distribution(cls, dist: pkg_resources.Distribution):
        if _is_dist_editable(dist):
            raise RuntimeError(f"Found editable installation which might have" f"local changes {dist.project_name}")
        req = dist.as_requirement()
        url = _get_distribution_url(dist)
        # if the URL is available, we'll remove the specifier (version info)
        # from the Requirement object since req has a bug when str(req) is
        # invoked for a req object with both version and URL. Nevertheless,
        # we wouldn't need the version info if url is present
        if url:
            req.url = url
            req.specifier = SpecifierSet("")
        # TODO: identify hashCmp, marker when possible
        return cls.from_requirement(req)

    def _version_check(self, other: PipPackage):
        """
        Checking the versions of two instances with special constraints.

        This fine grained check is triggered only in the case basic
        equality check (==) fails. It is for making sure version inclusion
        checks will pass. Specifically, version for package info fetched from
        environment would always has `==` as the version specifier since
        exact version is known. But version specifier for packages from
        requirements.txt could be any (`>` or `>=` or ...)

        Constraints:
            1. Number of specifier to the `self` must be 1. This is to
                make sure next constraint is valid
            2. The only specifier exist on `self` is `==`. This is to
                make sure the `self` knows exact version
            3. The version of `self` is included in the version of `other`
                Eg: `"pkg==1.2.0" == pkg>1.0.0,<=2.0.0

        Note: This check is based on the strong assumption that `other` is
        also a `PipPackage` but not a `CondaPackage`
        """
        self_has_version = bool(len(self._requirement_obj.specifier))
        other_has_version = bool(len(other._requirement_obj.specifier))
        if not self_has_version or not other_has_version:
            return NotImplemented
        if not bool(len(self._requirement_obj.specifier)):
            return NotImplemented
        if len(self._requirement_obj.specifier) > 1:
            # expecting other.__eq__ to solve the equality check
            return NotImplemented
        specifier = tuple(self._requirement_obj.specifier)[0]
        if specifier.operator != "==":
            # expecting other.__eq__ to solve the equality check
            return NotImplemented
        # checking at-least one specifier on the other side has
        # non-empty version
        for sp in other._requirement_obj.specifier:
            if sp.version:
                break
        else:
            return False
        # TODO: check for prereleases
        return specifier.version in other._requirement_obj.specifier

    def __str__(self):
        if self._requirement_obj:
            return str(self._requirement_obj)
        if self._original_string:
            return self._original_string
        raise RuntimeError("Found corrupted package")

    def __eq__(self, other: PipPackage):
        if self.__class__ != other.__class__:
            return NotImplemented

        if self.parsed_name != other.parsed_name:
            raise NonComparableError
        if self._requirement_obj is None or other._requirement_obj is None:
            # Comparison without the Requirement object needs to be
            # well thought out
            raise NonComparableError

        # Checking the equality on requirement object
        if self._requirement_obj == other._requirement_obj:
            return True
        # Checking the string representation since there are cases
        # Requirement's __eq__ check is not working as expected
        # pytorch-lightning-bolts installed from github is an example
        if str(self._requirement_obj) == str(other._requirement_obj):
            return True
        # Version inclusion check
        return self._version_check(other)


class DependencyManagerBase(abc.ABC):
    """Dependency manager base class.

    The dependency managers such as pip or conda or any future managers
    such as poetry should subclassed from this base class. The Abstract
    methods are specific to each package manager and should be implemented
    by the child class. These methods are being called in the setup time
    to build the information base by parsing the dependency listing and
    analyzing the environment file. It's also being utilized to write
    the dependency listing back to the package manager specific file
    """
    def __init__(self, dependency_file: Optional[str] = None):
        self._setup(dependency_file=dependency_file)

    @abc.abstractmethod
    def write_spec(self):
        """Write the dependency spec after formatting based on the dependency manager.
        """
        raise NotImplementedError

    @staticmethod
    @abc.abstractmethod
    def _get_default_req_file() -> Path:
        """Get the Path object for the requirement file for each dependency manager.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def _read_file_deps(self) -> Dict[str, BasePackage]:
        """Implement reading the requirement listing for each package manager.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def _fetch_env_deps(self) -> Dict[str, BasePackage]:
        """Implement reading dependencies from current environment.
        """
        raise NotImplementedError

    @property
    def source_deps(self):
        if not hasattr(self, "_source_deps_list"):
            raise RuntimeError("Manager is not ready. Run `_setup()`")
        if not isinstance(self._source_deps_list, list):
            raise RuntimeError("Source list is not a valid python list")
        return self._source_deps_list

    def _setup(self, dependency_file: Optional[str] = None):
        self._source_deps_list: List[str] = self._scan_source_deps()

        # Reading package manager specific methods to build information base
        if dependency_file:
            self.req_file = Path(dependency_file)
        else:
            self.req_file: Path = self._get_default_req_file()
        self.file_deps: Dict[str, BasePackage] = self._read_file_deps()
        self.env_deps: Dict[str, BasePackage] = {}
        for k, v in self._fetch_env_deps().items():
            if k in self._source_deps_list:
                self.env_deps[k] = v

    @property
    def has_change(self):
        """
        Return True if any change is needed. False other wise.
        """
        if not self.req_file.exists():
            return True
        if set(self._source_deps_list) - set(self.file_deps):
            return True
        for key in set(self._source_deps_list).intersection(self.env_deps):
            # TODO this equality check might have issues
            #  But the only way to make it work is to deploy in the wild
            #  and take feedback
            try:
                if self.env_deps[key] != self.file_deps[key]:
                    return True
            except NonComparableError:
                continue

    @staticmethod
    def _scan_source_deps() -> List[str]:
        # TODO: This could take a lot of time if files are a lot (1000s)
        """Scan source code in the CWD (recursively) and finds the
        package dependencies

        This information will then be used as the reference while fetching
        information from different sources such as current environment,
        requirement listing, pypi etc
        """
        ignore_dirs = ["venv", "test", "tests", "_test", "_tests", "egg", "EGG", "info", "docs", "__pycache__"]
        input_path = Path.cwd()
        for pth in input_path.iterdir():
            # ignore directories beginning with ('.' and '__')
            if pth.is_dir() and (pth.name.startswith('.') or pth.name.startswith('__')):
                ignore_dirs.append(pth.name)

        # TODO: remove pipreqs dependency
        # other considerations: pydeps & pigar
        candidates = pipreqs.get_all_imports(
            input_path, encoding=None, extra_ignore_dirs=ignore_dirs, follow_links=True
        )
        return [_source2parsed_pypi(c) for c in candidates]

    def get_missing(self) -> Collection[str]:
        """
        Get packages with proper information is missing. Any package which
        is found installed in environment and found specified in the
        requirement listing file (even if no version is specified), will
        be considered as "not missing". Other packages present in the source
        code will be returned as "missing"
        """
        available = set(self.env_deps) | set(self.file_deps)
        return set(self._source_deps_list) - available

    def collate_final(self):
        """
        Collate the final list of requirements for writing back to the
        disk.

        It loops through all the existing requirements in the requirement
        listing file and make sure the the spec (eg: version) hasn't been
        changed by comparing with the spec of installed package. It then
        loops through the packages those are missing in the requirement
        listing and add the spec by fetching it from the environment
        """
        if self.get_missing():
            raise ValueError("Found requirements without spec identified")
        collated = []
        # don't union on env_deps since it could have non-necessary packages.
        # That might be true for file_deps as well but that's something user
        # gave us explicitly and hence ignoring. Also, valid_names must keep
        # the order of values in requirements file

        missing = set(self._source_deps_list) - set(self.file_deps)
        for name in itertools.chain(self.file_deps.keys(), sorted(missing)):
            if name in self.env_deps and name in self.file_deps:
                try:
                    is_same = self.env_deps[name] == self.file_deps[name]
                    # Preferring file deps in case both are equal to avoid
                    # losing other information such as markers
                    if is_same:
                        qualified = self.file_deps[name]
                    else:
                        qualified = self.env_deps[name]
                except NonComparableError:
                    # Assuming both are same if not comparable
                    qualified = self.file_deps[name]
            elif name in self.file_deps:
                qualified = self.file_deps[name]
            elif name in self.env_deps:
                qualified = self.env_deps[name]
            else:
                raise RuntimeError(f"Package name '{name}' not found in file " f"or env")
            collated.append(qualified)
        return collated

    def write_config(self, config_path: Optional[str] = None):
        names = set()
        for n in self._source_deps_list:
            names |= set(_get_system_deps(n))
        if names:
            config_path = Path(config_path or "config.yml")
            if config_path.exists():
                from grid.cli.cli.utilities import read_config
                config = read_config(value=config_path.open('r'))
            else:
                config = {"compute": {}}
                config_path.touch()
            # Sorting to make the order deterministic
            sys_deps_action = f"apt install -y {' '.join(sorted(names))}"
            if "actions" not in config["compute"]:
                config["compute"]["actions"] = {"on_image_build": []}
            elif "on_image_build" not in config["compute"]["actions"]:
                config["compute"]["actions"]["on_image_build"] = []

            if not isinstance(config["compute"]["actions"]["on_image_build"], list):
                raise RuntimeError("Parsing error while reading config.yml")

            config["compute"]["actions"]["on_image_build"].append(sys_deps_action)
            yaml.safe_dump(config, config_path.open(mode="w+"))


class CondaManager(DependencyManagerBase):
    _env_name: Optional[str] = None
    _prefix: Optional[str] = None
    _channels: List[str] = []

    @staticmethod
    def _build_dependency_dict(specs: List) -> Dict[str, Union[CondaPackage, PipPackage]]:
        """
        Build dependency objects from the spec

        Parameters
        ----------
        specs:
            Dependency specs read from the environment.yml file. It would
            contain the all the dependencies formatted as "name=version=build".
            It would also contain PIP dependencies as a dictionary with
            key as "pip". An example `specs` is given below

            specs = ["numpy-base=1.19.2=py38hcfb5961_0",
                     "ninja=1.10.2=py38hf7b0b51_0",
                     ...
                     {"pip": ["absl-py==0.11.0"
                              "attrs==20.2.0",
                              ...
                             ]
                     }
                     ]

            We are not expecting it to contain a dictionary with key other than
            "pip" and hence we throw if that's the case

        Returns
        -------
        A dictionary that maps the parsed pypi name of each dependency to the
        Package object. Since conda dependency listing has both conda and
        pip spec, the return value would contain both CondaPackage and
        PipPackage containers
        """
        dependencies = {}
        pip_dependencies = []
        for d in specs:
            # Only pip listing inside a yaml file must be a mapping
            if isinstance(d, dict):
                if "pip" not in d or len(d) > 1:
                    raise RuntimeError("Could not parse the conda yaml file. " "Raise the issue with Grid support")
                pip_dependencies = d["pip"]
            # conda deps listing in a yaml file will be a string
            else:
                try:
                    dep = CondaPackage.from_string(d)
                except RuntimeError:
                    raise RuntimeError("Could not parse the conda yaml file. " "Raise the issue with Grid support")
                dependencies[dep.parsed_name] = dep
        for elem in pip_dependencies:
            dep = PipPackage.from_string(elem)
            dependencies[dep.parsed_name] = dep
        return dependencies

    def _read_file_deps(self) -> Dict[str, CondaPackage]:
        """
        Read environment.yml to fetch user defined dependencies
        """
        if self.req_file.exists():
            specs = yaml.safe_load(self.req_file.read_text())
            if not isinstance(specs, dict):
                raise RuntimeError(f"Parsing error with {self.req_file}")
            self._env_name = specs.get("name")
            self._prefix = specs.get("prefix")
            self._channels = self._channels + specs["channels"]
            return self._build_dependency_dict(specs["dependencies"])
        return {}

    @staticmethod
    def _get_default_req_file() -> Path:
        return Path("environment.yml")

    def _fetch_env_deps(self) -> Dict[str, CondaPackage]:
        """
        Fetch dependencies from currently active conda environment
        """
        if not os.getenv("CONDA_DEFAULT_ENV"):
            return {}
        export = execute_conda_command(["env", "export"])
        specs = yaml.safe_load(export)

        self._env_name = self._env_name or specs.get("name")
        self._channels = self._channels + specs["channels"]
        return self._build_dependency_dict(specs["dependencies"])

    def write_spec(self):
        """
        Write requirements to `environment.yml`
        """
        final = {}
        if self._env_name:
            final["name"] = self._env_name
        if self._prefix:
            final["prefix"] = self._prefix
        final["channels"] = sorted(list(set(self._channels)))
        conda_deps = []
        pip_deps = []
        collated = self.collate_final()
        for dep in collated:
            if isinstance(dep, PipPackage):
                pip_deps.append(str(dep))
            else:
                conda_deps.append(str(dep))
        final["dependencies"] = conda_deps
        final["dependencies"].append({"pip": pip_deps})  # noqa
        yaml.safe_dump(final, self.req_file.open(mode="w+"))


def _yield_valid_lines(content):
    """Yield valid lines from requirements.txt
    """
    lines = (line for line in content.splitlines())
    for line in lines:
        line = line.strip()
        if not line or line.startswith("#"):
            continue
        # Drop comments but not # in a URL
        if ' #' in line:
            line = line[:line.find(' #')]
        # line continuations
        if line.endswith('\\'):
            line = line[:-2].strip()
            try:
                line += next(lines)
            except StopIteration:
                return
        yield line


class PipManager(DependencyManagerBase):
    def _read_file_deps(self) -> Dict[str, PipPackage]:
        """
        Read requirements.txt to fetch user defined dependencies
        """
        file_deps = {}
        if self.req_file.exists():
            for ln in _yield_valid_lines(self.req_file.read_text()):
                dep = PipPackage.from_string(ln)
                file_deps[dep.parsed_name] = dep
        return file_deps

    @staticmethod
    def _get_default_req_file() -> Path:
        return Path("requirements.txt")

    def _fetch_env_deps(self) -> Dict[str, PipPackage]:
        """
        Look at the currently active environment to fetch the list of installed
        dependencies.
        """
        out = {}
        working_set = pkg_resources.working_set
        if working_set:
            for ws in working_set:  # skipcq: PYL-E1133
                dep = PipPackage.from_distribution(ws)
                out[dep.parsed_name] = dep
        return out

    def write_spec(self):
        """
        Write requirements to `requirements.txt`
        """
        collated = self.collate_final()
        out = "\n".join((str(val) for val in collated)) + "\n"
        Path("requirements.txt").write_text(out)



# coding: utf-8

"""
    external/v1/external_session_service.proto

    No description provided (generated by Swagger Codegen https://github.com/swagger-api/swagger-codegen)  # noqa: E501

    OpenAPI spec version: version not set
    
    Generated by: https://github.com/swagger-api/swagger-codegen.git

    NOTE
    ----
    standard swagger-codegen-cli for this python client has been modified
    by custom templates. The purpose of these templates is to include
    typing information in the API and Model code. Please refer to the
    main grid repository for more info
"""


import pprint
import re  # noqa: F401
from typing import TYPE_CHECKING

import six

from grid.openapi.configuration import Configuration

if TYPE_CHECKING:
    from datetime import datetime
    from grid.openapi.models import *

class V1ExperimentCounts(object):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    """
    Attributes:
      swagger_types (dict): The key is attribute name
                            and the value is attribute type.
      attribute_map (dict): The key is attribute name
                            and the value is json key in definition.
    """
    swagger_types = {
        'cancelled': 'int',
        'deleted': 'int',
        'failed': 'int',
        'pending': 'int',
        'running': 'int',
        'succeeded': 'int'
    }

    attribute_map = {
        'cancelled': 'cancelled',
        'deleted': 'deleted',
        'failed': 'failed',
        'pending': 'pending',
        'running': 'running',
        'succeeded': 'succeeded'
    }

    def __init__(self, cancelled: 'int' = None, deleted: 'int' = None, failed: 'int' = None, pending: 'int' = None, running: 'int' = None, succeeded: 'int' = None, _configuration=None):  # noqa: E501
        """V1ExperimentCounts - a model defined in Swagger"""  # noqa: E501
        if _configuration is None:
            _configuration = Configuration()
        self._configuration = _configuration

        self._cancelled = None
        self._deleted = None
        self._failed = None
        self._pending = None
        self._running = None
        self._succeeded = None
        self.discriminator = None

        if cancelled is not None:
            self.cancelled = cancelled
        if deleted is not None:
            self.deleted = deleted
        if failed is not None:
            self.failed = failed
        if pending is not None:
            self.pending = pending
        if running is not None:
            self.running = running
        if succeeded is not None:
            self.succeeded = succeeded

    @property
    def cancelled(self) -> 'int':
        """Gets the cancelled of this V1ExperimentCounts.  # noqa: E501


        :return: The cancelled of this V1ExperimentCounts.  # noqa: E501
        :rtype: int
        """
        return self._cancelled

    @cancelled.setter
    def cancelled(self, cancelled: 'int'):
        """Sets the cancelled of this V1ExperimentCounts.


        :param cancelled: The cancelled of this V1ExperimentCounts.  # noqa: E501
        :type: int
        """

        self._cancelled = cancelled

    @property
    def deleted(self) -> 'int':
        """Gets the deleted of this V1ExperimentCounts.  # noqa: E501


        :return: The deleted of this V1ExperimentCounts.  # noqa: E501
        :rtype: int
        """
        return self._deleted

    @deleted.setter
    def deleted(self, deleted: 'int'):
        """Sets the deleted of this V1ExperimentCounts.


        :param deleted: The deleted of this V1ExperimentCounts.  # noqa: E501
        :type: int
        """

        self._deleted = deleted

    @property
    def failed(self) -> 'int':
        """Gets the failed of this V1ExperimentCounts.  # noqa: E501


        :return: The failed of this V1ExperimentCounts.  # noqa: E501
        :rtype: int
        """
        return self._failed

    @failed.setter
    def failed(self, failed: 'int'):
        """Sets the failed of this V1ExperimentCounts.


        :param failed: The failed of this V1ExperimentCounts.  # noqa: E501
        :type: int
        """

        self._failed = failed

    @property
    def pending(self) -> 'int':
        """Gets the pending of this V1ExperimentCounts.  # noqa: E501


        :return: The pending of this V1ExperimentCounts.  # noqa: E501
        :rtype: int
        """
        return self._pending

    @pending.setter
    def pending(self, pending: 'int'):
        """Sets the pending of this V1ExperimentCounts.


        :param pending: The pending of this V1ExperimentCounts.  # noqa: E501
        :type: int
        """

        self._pending = pending

    @property
    def running(self) -> 'int':
        """Gets the running of this V1ExperimentCounts.  # noqa: E501


        :return: The running of this V1ExperimentCounts.  # noqa: E501
        :rtype: int
        """
        return self._running

    @running.setter
    def running(self, running: 'int'):
        """Sets the running of this V1ExperimentCounts.


        :param running: The running of this V1ExperimentCounts.  # noqa: E501
        :type: int
        """

        self._running = running

    @property
    def succeeded(self) -> 'int':
        """Gets the succeeded of this V1ExperimentCounts.  # noqa: E501


        :return: The succeeded of this V1ExperimentCounts.  # noqa: E501
        :rtype: int
        """
        return self._succeeded

    @succeeded.setter
    def succeeded(self, succeeded: 'int'):
        """Sets the succeeded of this V1ExperimentCounts.


        :param succeeded: The succeeded of this V1ExperimentCounts.  # noqa: E501
        :type: int
        """

        self._succeeded = succeeded

    def to_dict(self) -> dict:
        """Returns the model properties as a dict"""
        result = {}

        for attr, _ in six.iteritems(self.swagger_types):
            value = getattr(self, attr)
            if isinstance(value, list):
                result[attr] = list(map(
                    lambda x: x.to_dict() if hasattr(x, "to_dict") else x,
                    value
                ))
            elif hasattr(value, "to_dict"):
                result[attr] = value.to_dict()
            elif isinstance(value, dict):
                result[attr] = dict(map(
                    lambda item: (item[0], item[1].to_dict())
                    if hasattr(item[1], "to_dict") else item,
                    value.items()
                ))
            else:
                result[attr] = value
        if issubclass(V1ExperimentCounts, dict):
            for key, value in self.items():
                result[key] = value

        return result

    def to_str(self) -> str:
        """Returns the string representation of the model"""
        return pprint.pformat(self.to_dict())

    def __repr__(self) -> str:
        """For `print` and `pprint`"""
        return self.to_str()

    def __eq__(self, other: 'V1ExperimentCounts') -> bool:
        """Returns true if both objects are equal"""
        if not isinstance(other, V1ExperimentCounts):
            return False

        return self.to_dict() == other.to_dict()

    def __ne__(self, other: 'V1ExperimentCounts') -> bool:
        """Returns true if both objects are not equal"""
        if not isinstance(other, V1ExperimentCounts):
            return True

        return self.to_dict() != other.to_dict()



from typing import List, Optional

from grid.metadata import __version__
from grid.sdk.client import create_gql_client
from grid.sdk.client.grid_gql import gql_execute


def get_user_basic_info():
    client = create_gql_client()
    query = """
    query {
        getUser {
            userId
            isVerified
            completedSignup
            isBlocked
            username
            firstName
            lastName
            email
        }
    }
    """
    return gql_execute(client, query)['getUser']


def get_user_teams() -> List[dict]:
    client = create_gql_client()
    query = """
        query GetUserTeams {
            getUserTeams {
                success
                message
                teams {
                    id
                    name
                    createdAt
                    role
                    members {
                        id
                        username
                        firstName
                        lastName
                    }
                }
            }
        }
    """
    result = gql_execute(client, query)
    if not result['getUserTeams'] or not result['getUserTeams']['success']:
        raise RuntimeError(result['getUserTeams']["message"])
    return result['getUserTeams']['teams']


def get_user_info():
    """Return basic information about a user."""
    client = create_gql_client()
    query = """
        query {
            getUser {
                username
                firstName
                lastName
                email

            }
        }
    """

    result = gql_execute(client, query)
    if not result['getUser']:
        raise RuntimeError(result['getUser']["message"])
    return result['getUser']



import contextlib
import textwrap
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Iterable, Iterator, List, Optional, Union, Dict
from re import compile

from grid.sdk import env
from grid.sdk.artifacts import Artifact
from grid.sdk.affirmations import affirm, is_not_deleted, is_created, is_not_created, is_not_shallow
from grid.sdk.run_components import Resources, Actions, ScratchSpace
from grid.sdk.utilities import (
    read_config,
    check_run_name_is_valid,
    check_description_isnt_too_long,
    resolve_instance_type_nickname,
    SPECIAL_NAME_TO_SKIP_OBJECT_INIT,
)
from grid.sdk.client import create_swagger_client
from grid.sdk.datastores import Datastore
from grid.sdk.experiments import Experiment
from grid.sdk.rest import GridRestClient
from grid.sdk.rest.runs import create_run, get_run_from_name, cancel_run, delete_run, get_run_from_id
from grid.sdk.rest.runs import list_runs as rest_list_runs
from grid.sdk.utils import fill_object_from_response
from grid.sdk.source_code.repository import GitRepository, LocalSourceCodeDir
from grid.sdk.user import user_from_logged_in_account, User
from grid.sdk.utils.name_generator import unique_name

from grid.openapi import V1Run, V1RunState


def check_valid_framework(framework) -> bool:
    return bool(
        framework in [
            'lightning',
            'torch',
            'pytorch',
            'tensorflow',
            'julia',
            'julia:1.6.1',
            'julia:1.6.2',
            'julia:1.6.3',
            'julia:1.6.4',
            'julia:1.6.5',
            'julia:1.7.0',
            'julia:1.7.1',
            'torchelastic',
        ]
    )


def validate_random_search_strategy(strategy_options: Dict[str, str]):
    print(strategy_options)
    """ checks that strategy for random search will not fail later on """
    if strategy_options is None:
        raise ValueError('Strategy "random_search" requires at least the "num_trials" option to use.')
    num_trials = strategy_options.get('num_trials', '')
    if num_trials == '':
        raise ValueError('A value for `num_trials` is required when using the `random_search` sweep strategy.')
    try:
        num_trials_int = int(num_trials)
    except ValueError:
        raise ValueError(f'A value for `num_trials` should be an integer but we got "{num_trials}".')
    else:
        if num_trials_int <= 0:
            raise ValueError(f'A value for `num_trials` should be > 0. We got "{num_trials_int}".')


class Run:
    """A ``Run`` is a container for a set of computations executed over search space.

    This object can be instantiated with constructor args below in
    order to create/start a new set of computations (``Experiments``)
    which make up the ``Run``.

    Alternatively, an existig ``Run`` ``name`` can be passed in the
    constructor, and the SDK will load all the associated ``Run`` /
    ``Experiment`` details in this structure automatically.

    !!! note
        This class provides properties to access run attributes
        which can be read at any time, or set before the ``Run``
        objects ``.start()`` method is called.
    """
    _user: User
    _id: str

    _name: str
    _run_command: str
    _url: str
    _path: Path
    _description: str
    _strategy: str
    _strategy_options: Dict[str, str]
    _env_vars: Dict[str, str]
    _framework: str
    _dependency_file: Union[str, Path]
    _localdir: bool
    _relative_working_dir: Path
    _dockerfile: Union[str, Path]
    _auto_resume: bool
    _resources: Resources
    _datastore: Optional[Datastore]
    _datastore_mount_dir: Optional[str]
    _actions: Actions
    _scratch: ScratchSpace
    _cluster_id: str
    _dry_run: bool
    _status: 'RunState'

    _cost: Union[int, float]
    _estimated_hourly_cost: Union[int, float]
    _desired_status: 'RunState'
    _created_at: datetime
    _updated_at: datetime
    _experiments: List['Experiment']
    _experiment_counts: Dict[str, int]

    _client: GridRestClient

    _is_deleted: bool  # TODO use _status instead
    _is_created: bool
    _is_shallow: bool

    def __init__(
        self,
        name: Optional[str] = None,
        run_command: Optional[str] = None,
        description: Optional[str] = None,
        strategy: Optional[str] = "grid_search",
        strategy_options: Optional[Dict[str, str]] = None,  # if strategy == random_search
        env_vars: Optional[Dict[str, str]] = None,
        framework: Optional[str] = "lightning",
        dependency_file: Optional[Union[str, Path]] = "./requirements.txt",
        localdir: Optional[bool] = False,
        url: Optional[str] = None,
        path: Optional[Union[str, Path]] = None,
        dockerfile: Optional[Union[str, Path]] = None,
        auto_resume: Optional[bool] = False,
        resources: Resources = None,
        datastore: Optional[Datastore] = None,
        datastore_mount_dir: Optional[str] = None,
        actions: Optional[Actions] = None,
        scratch: Optional[ScratchSpace] = None,
        cluster_id: Optional[str] = None,
        config_file: Optional[Union[str, Path]] = None,
        dry_run: bool = False
    ):

        # --------------------------------------------------------------------------- #
        #    This should be the first block that goes into the constructor of the     #
        #    resource object. This block sets the correct values for the private      #
        #    attributes which is then later picked up by the decorator(s) to take     #
        #    right actions. It also initialize the _client object and cluster_id      #
        #    which is required by the downstream methods regardless of the object     #
        #    is completely initialized or not. Most importantly, this blocks checks   #
        #    for the name argument to decide if it's a call from other internal       #
        #    methods to create a shallow object. This is done by checking the         #
        #    special name variable. Other methods that already has the backend        #
        #    response fetched, can use this to create the object without the backend  #
        #    call and then fill-in the response they already have.                    #
        #                                                                             #
        self._client = GridRestClient(api_client=create_swagger_client())
        cluster_id = cluster_id or env.CONTEXT
        self._is_shallow = False
        self._cluster_id = cluster_id
        if name == SPECIAL_NAME_TO_SKIP_OBJECT_INIT:
            self._is_shallow = True
            self._is_created = False
            self._is_deleted = False
            return
        #                                                                             #
        # --------------------------------------------------------------------------- #

        if name is None:
            name = unique_name()
        else:
            # if user passed a name, try to fetch the run from BE
            try:
                run = get_run_from_name(client=self._client, cluster_id=cluster_id, run_name=name)
                self._setup_from_response(run)
                return
            except KeyError:
                self._is_deleted = False  # the run has not been deleted
                self._is_created = False  # it doesn't exists in the grid backend.
                pass
        check_run_name_is_valid(name)
        self._name = name
        if description:
            check_description_isnt_too_long(description)
        if resources is None:
            resources = Resources()
        if scratch is None:
            scratch = ScratchSpace()
        if env_vars is None:
            env_vars = {}
        if actions is None:
            actions = Actions()

        if resources.instance_type:
            resolve_instance_type_nickname(resources.instance_type)

        if config_file:
            config = read_config(config_file)
            # from provider sub-key.
            cluster_id = config.get('compute', {}).get('provider', {}).get('cluster', cluster_id)

            # from config/hyper_params sub-key.
            _hparam_cfg = config.get("hyper_params", {})
            strategy = _hparam_cfg.get('settings', {}).get('strategy', strategy)
            strategy_options = _hparam_cfg.get("settings", {}).get("strategy_options", strategy_options)
            # if not already provided update with config file values (note there is no complementing from config value to pass values)
            if strategy == 'random_search' and strategy_options is None:
                strategy_options = {
                    'num_trials': _hparam_cfg.get('settings', {}).get('num_trials', None),
                    'seed': _hparam_cfg.get('settings', {}).get('seed', '0')
                }

            _params = _hparam_cfg.get('params', {})
            if not isinstance(_params, dict):
                raise ValueError("params should be a key value pair mapping.")
            else:
                if _params:
                    for key in _params:
                        run_command += f' {key} {_params[key]}'

            # from config/compute/train sub-key
            _train_cfg = config.get('compute', {}).get('train', {})
            env_vars = _train_cfg.get("environment", env_vars)
            dry_run = _train_cfg.get('dry_run', dry_run)
            framework = _train_cfg.get('framework', framework)
            dependency_file = _train_cfg.get('dependency_file_info', {}).get('path', dependency_file)
            resources.instance_type = _train_cfg.get('instance', resources.instance_type)
            resources.use_spot = _train_cfg.get('use_spot', resources.use_spot)
            resources.cpus = _train_cfg.get('cpus', resources.cpus)
            resources.gpus = _train_cfg.get('gpus', resources.gpus)
            resources.storage_gb = _train_cfg.get('storage_gb', resources.storage_gb)
            localdir = _train_cfg.get('localdir', localdir)
            dockerfile = _train_cfg.get('dockerfile', dockerfile)
            auto_resume = _train_cfg.get('auto_resume', auto_resume)

            _dstore_config = [_train_cfg.get('datastore_name'), _train_cfg.get('datastore_version')]
            if any(_dstore_config) and not all(_dstore_config):
                raise ValueError(
                    "name and version are both required if one of datastore_name or datastore_version is set"
                )
            if all(_dstore_config):
                _dstore_name = _train_cfg.get('datastore_name')
                _dstore_version = _train_cfg.get('datastore_version')
                datastore_mount_dir = _train_cfg.get('datastore_mount_dir', datastore_mount_dir)
                datastore = Datastore(name=_dstore_name, version=_dstore_version, cluster_id=cluster_id)

            _actions_cfg = _train_cfg.get("actions", {})
            if len(_actions_cfg) != 0:
                actions = Actions(
                    on_build=_actions_cfg.get("on_build", actions.on_build),
                    on_build_start=_actions_cfg.get("on_build_start", actions.on_build_start),
                    on_build_end=_actions_cfg.get("on_build_end", actions.on_build_end),
                    on_experiment_start=_actions_cfg.get("on_experiment_start", actions.on_experiment_start),
                    on_experiment_end=_actions_cfg.get("on_experiment_end", actions.on_experiment_end),
                )
            _scratch = _train_cfg.get("scratch", [])
            if len(_scratch) != 0:
                if len(_scratch) > 1:
                    raise ValueError("Only one scratch space is allowed")
                scratch = ScratchSpace(
                    size_gb=_scratch.get("size_gb", scratch.size_gb),
                    mount_path=_scratch.get("mount_path", scratch.mount_path),
                )

        if not framework or not check_valid_framework(framework):
            raise ValueError(f'framework: {framework} is not valid')

        if resources.storage_gb is not None and resources.storage_gb < 100:
            raise ValueError("Invalid disk size, should be greater than 100Gb")

        relative_working_dir = None
        if not url:
            path = Path(path) if path else Path.cwd()
            path = path.absolute()
            if dockerfile:
                with contextlib.suppress(ValueError):
                    dockerfile = str(Path(dockerfile).relative_to(path))
            if dependency_file:
                with contextlib.suppress(ValueError):
                    dependency_file = str(Path(dependency_file).relative_to(path))
            if not localdir:
                # in the case of localdir, we'll upload the code to the URL
                # server returns on run creation
                git_repo = GitRepository(path)
                url = git_repo.url()
                relative_working_dir = git_repo.relative_working_dir()

        if not datastore_mount_dir and isinstance(datastore, Datastore):
            datastore_mount_dir = f'/datastores/{datastore.name}'

        if strategy == 'random_search':
            validate_random_search_strategy(strategy_options)

        self._user = None
        self._cluster_id = cluster_id
        self._url = url
        self._path = path
        self._relative_working_dir = relative_working_dir
        self._env_vars = env_vars
        self._run_command = run_command
        self._description = description
        self._strategy = strategy
        self._strategy_options = strategy_options
        self._framework = framework if framework != 'torch' else 'pytorch'
        self._dependency_file = dependency_file
        self._localdir = localdir
        self._dockerfile = dockerfile
        self._auto_resume = auto_resume
        self._resources = resources
        self._datastore = datastore
        self._datastore_mount_dir = datastore_mount_dir
        self._actions = actions
        self._scratch = scratch
        self._dry_run = dry_run

        self._id = None
        self._desired_status = None
        self._created_at = None
        self._updated_at = None
        self._status = None
        self._experiments = []
        self._experiment_counts = None
        self._cost = None
        self._estimated_hourly_cost = None

    @property
    @affirm(is_not_shallow)
    def exists(self) -> bool:
        return self._is_created

    @classmethod
    def _from_id(cls, run_id: str, cluster_id: Optional[str] = env.CONTEXT) -> "Run":
        instance = cls(name=SPECIAL_NAME_TO_SKIP_OBJECT_INIT, cluster_id=cluster_id)
        instance._id = run_id
        instance._is_shallow = True
        return instance

    def _setup_from_response(self, run: 'V1Run'):
        """Set up the run from the response from the API."""
        self._is_created = True
        self._is_deleted = run.status.phase == V1RunState.DELETED
        self._is_shallow = False

        self._name = run.name
        self._id = run.id
        self._dry_run = run.spec.dry_run

        self._run_command = run.spec.run_controller_command
        self._description = run.description
        self._cluster_id = run.spec.cluster_id
        self._strategy = run.spec.sweep_type
        self._strategy_options = run.spec.sweep_options
        self._framework = run.spec.image.framework
        self._dockerfile = run.spec.image.dockerfile
        self._localdir = bool('s3:' in run.spec.source_code)

        if run.spec.image.dependency_file_info:
            self._dependency_file = run.spec.image.dependency_file_info.path

        self._desired_status = RunState.from_api_spec(run.spec.desired_state)
        self._status = RunState.from_api_spec(run.status.phase)
        self._created_at = run.created_at
        self._updated_at = run.updated_at

        self._experiment_counts = {}
        if run.status.experiment_counts:
            self._experiment_counts = run.status.experiment_counts.to_dict()

        self._experiments = []
        for exp_id in run.status.experiment_ids:
            try:
                # noinspection PyProtectedMember
                exp = Experiment._from_id(cluster_id=run.spec.cluster_id, exp_id=exp_id)
                self._experiments.append(exp)
            except Exception as e:
                # Experiment might be deleted
                if not hasattr(e, 'reason') or e.reason != 'Not Found':  # noqa
                    raise e
        storage_gb = int(run.spec.resources.storage_gb) if run.spec.resources.storage_gb else None
        cpus = int(run.spec.resources.cpu) if run.spec.resources.cpu else None
        gpus = int(run.spec.resources.gpu) if run.spec.resources.gpu else None
        self._resources = Resources(
            instance_type=run.spec.instance_type,
            use_spot=run.spec.use_spot,
            storage_gb=storage_gb,
            cpus=cpus,
            gpus=gpus,
            extra=run.spec.resources.extra,
        )

        self._datastore = None
        self._datastore_mount_dir = None
        for datastore_mount in run.spec.datastores:
            # noinspection PyProtectedMember
            self._datastore = Datastore._from_id(cluster_id=run.spec.cluster_id, datastore_id=datastore_mount.id)
            self._datastore_mount_dir = datastore_mount.mount_path

        if len(run.spec.scratch) == 1:
            scratch = run.spec.scratch[0]
            self._scratch = ScratchSpace(size_gb=scratch.size_gb, mount_path=scratch.mount_path)
        if len(run.spec.scratch) > 1:
            raise ValueError(f"Expected only one scratch space, but got {len(run.spec.scratch)}")

        self._cost = run.status.cost
        self._estimated_hourly_cost = run.status.hourly_cost
        self._user = User(user_id=run.spec.user_id, username="", first_name="", last_name="")

    @property
    def id(self) -> str:
        return self._id

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def name(self) -> str:
        """Name of the Run

        Returns
        -------
        str
            The specified (or autogenerated) Run name.
        """
        return self._name

    @name.setter
    @affirm(is_not_deleted, is_not_created)
    def name(self, val: str):
        self._name = val

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def run_command(self) -> str:
        """The command you would you to start the main script from the root of the source tree.

        For example:

            python scripts/start_foo.py --arg "[1, 2]" --flag
        """
        return self._run_command

    @run_command.setter
    @affirm(is_not_deleted, is_not_created)
    def run_command(self, val: str):
        self._run_command = val

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def url(self) -> str:
        """Remote URL to the repository
        """
        return self._url

    @url.setter
    @affirm(is_not_deleted, is_not_created)
    def url(self, val: str):
        self._url = val

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def relative_working_dir(self) -> Path:
        """local path to the repository
        """
        return self._relative_working_dir

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def path(self) -> Path:
        """local path to the repository
        """
        return self._path

    @path.setter
    @affirm(is_not_deleted, is_not_created)
    def path(self, val: Path):
        self._path = val

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def description(self) -> str:
        """Human readable description of the Run's contents

        Returns
        -------
        str
            A description string if one was attached to the Run.
        """
        return self._description

    @description.setter
    @affirm(is_not_deleted, is_not_created)
    def description(self, val: str):
        self._description = val

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def strategy(self) -> str:
        """Hyperparameter search strategy.

        Returns
        -------
        str
            One of `['grid_search', 'random_search', 'none']`
        """
        return self._strategy

    @strategy.setter
    @affirm(is_not_deleted, is_not_created)
    def strategy(self, val: str):
        self._strategy = val

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def strategy_options(self) -> Dict[str, str]:
        """Returns the options for the `random_search` sweep strategy.

        Returns
        -------
        Dict[str, str]
            There are two valid keys: `num_trials` and `seed`:
                num_trials:   how many samples of the full search space are used
                seed:         seed to initialize the random_search
        """
        return self._strategy_options

    @strategy_options.setter
    @affirm(is_not_deleted, is_not_created)
    def strategy_options(self, val: Dict[str, str]):
        """Set the options for the `random_search` sweep strategy.

        Parameters
        ----------
        val : Dict[str, str]
            There are two valid keys: `num_trials` and `seed`:
                num_trials:   how many samples of the full search space are used
                seed:         seed to initialize the random_search
        """
        self._strategy_options = val

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def env_vars(self) -> Dict[str, str]:
        return self._env_vars

    @env_vars.setter
    @affirm(is_not_deleted, is_not_created)
    def env_vars(self, val: Dict[str, str]):
        self._env_vars = val

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def framework(self) -> str:
        """Determines the ML framework used.

        Returns
        -------
        str
            One of ``["lightning", "pytorch", "tensorflow", "julia"]``.
        """
        return self._framework

    @framework.setter
    @affirm(is_not_deleted, is_not_created)
    def framework(self, val: str):
        self._framework = val

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def dependency_file(self) -> Union[str, Path]:
        """Path to a dependency file to use to install runtime requirements.

        Returns
        -------
        Union[str, Path]
            Path to a dependency file.
        """
        return self._dependency_file

    @dependency_file.setter
    @affirm(is_not_deleted, is_not_created)
    def dependency_file(self, val: Union[str, Path]):
        self._dependency_file = val

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def localdir(self) -> bool:
        return self._localdir

    @localdir.setter
    @affirm(is_not_deleted, is_not_created)
    def localdir(self, val: bool):
        self._localdir = val

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def dockerfile(self) -> Union[str, Path]:
        return self._dockerfile

    @dockerfile.setter
    @affirm(is_not_deleted, is_not_created)
    def dockerfile(self, val: Union[str, Path]):
        self._dockerfile = val

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def auto_resume(self) -> bool:
        return self._auto_resume

    @auto_resume.setter
    @affirm(is_not_deleted, is_not_created)
    def auto_resume(self, val: bool):
        self._auto_resume = val

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def resources(self) -> Resources:
        return self._resources

    @resources.setter
    @affirm(is_not_deleted, is_not_created)
    def resources(self, val: Resources):
        self._resources = val

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def datastore(self) -> Datastore:
        return self._datastore

    @datastore.setter
    @affirm(is_not_deleted, is_not_created)
    def datastore(self, val: Datastore):
        self._datastore = val

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def datastore_mount_dir(self) -> str:
        return self._datastore_mount_dir

    @datastore_mount_dir.setter
    @affirm(is_not_deleted, is_not_created)
    def datastore_mount_dir(self, val: str):
        self._datastore_mount_dir = val

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def actions(self) -> Actions:
        return self._actions

    @actions.setter
    @affirm(is_not_deleted, is_not_created)
    def actions(self, val: Actions):
        self._actions = val

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def scratch(self) -> ScratchSpace:
        return self._scratch

    @scratch.setter
    @affirm(is_not_deleted, is_not_created)
    def scratch(self, val: ScratchSpace):
        self._scratch = val

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def cluster_id(self) -> str:
        return self._cluster_id

    @cluster_id.setter
    @affirm(is_not_deleted, is_not_created)
    def cluster_id(self, cluster: str):
        self._cluster_id = cluster

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def dry_run(self) -> bool:
        return self._dry_run

    @dry_run.setter
    @affirm(is_not_deleted, is_not_created)
    def dry_run(self, dry_run: bool):
        self._dry_run = dry_run

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def status(self) -> 'RunState':
        return self._status

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def desired_status(self) -> 'RunState':
        return self._desired_status

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def created_at(self) -> datetime:
        return self._created_at

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def updated_at(self) -> datetime:
        return self._updated_at

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def experiments(self) -> List['Experiment']:
        return self._experiments

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def experiment_counts(self) -> Dict[str, int]:
        return self._experiment_counts

    @property
    @affirm(is_not_shallow, is_not_deleted)
    def user(self) -> User:
        if self._user is None:
            self._user = user_from_logged_in_account()
        return self._user

    @property
    def cost(self) -> Union[int, float]:
        return self._cost

    @property
    def estimated_hourly_cost(self) -> Union[int, float]:
        """Estimated hourly cost of the run.

        Returns
        -------
        Union[int, float]
            Value corresponding to (estimated) cost / hr in US Dollars.

        TODO(rlizzo): Update to use the "dry-run" API to get the estimated cost.
        """
        return self._estimated_hourly_cost

    @affirm(is_not_created)
    def start(self) -> bool:
        """Launches the hyper-parameter sweep computations on the cluster.

        After starting a Run, most of the run configuration cannot be changed.
        If successful, the machine resources requested will be provisioned and
        begin billing your account until the run completes, fails, is cancelled,'
        or your account runs out of funds.
        """
        if not self.cluster_id:
            raise ValueError("Couldn't find cluster info. Try logging in again!")
        datastores = {self.datastore.id: self.datastore_mount_dir} if self.datastore else None
        res = create_run(
            client=self._client,
            cluster_id=self.cluster_id,
            run_name=self.name,
            run_description=self.description,
            run_controller_command=self.run_command,
            source_code=self.url,
            localdir=self.localdir,
            run_sweep_type=self.strategy,
            run_sweep_options=self.strategy_options,
            scratch_mount_path=self.scratch.mount_path,
            scratch_size_gb=self.scratch.size_gb,
            instance_type=self.resources.instance_type,
            use_spot=self.resources.use_spot,
            per_exp_resources_cpu=self.resources.cpus,
            per_exp_resources_gpu=self.resources.gpus,
            per_exp_resources_storage_gb=self.resources.storage_gb,
            per_exp_resources_extra=self.resources.extra,
            on_build_actions=self.actions.on_build,
            on_build_start_actions=self.actions.on_build_start,
            on_build_end_actions=self.actions.on_build_end,
            on_experiment_start_actions=self.actions.on_experiment_start,
            on_experiment_end_actions=self.actions.on_experiment_end,
            image_dockerfile=self.dockerfile,
            image_framework=self.framework,
            image_dep_file=self.dependency_file,
            per_exp_env_vars=self.env_vars,
            datastores=datastores,
            relative_work_dir=str(self.relative_working_dir) if self.relative_working_dir else None,
            dry_run=False,  # TODO
            auto_resume=self.auto_resume,
        )

        # trigger the repo upload if localdir is true
        if self.localdir:
            if not self.path.is_dir():
                raise NotADirectoryError(f'localdir_source_root: {self.path} is not a directory')
            if res.status.phase == V1RunState.FAILED:
                raise RuntimeError(f'Run creation failed: {res.status.message}')
            repository = LocalSourceCodeDir(path=self.path)
            repository.package()
            if not res.status.upload_url:
                raise ValueError("Couldn't get upload url for localdir source")
            repository.upload(url=res.status.upload_url)

        self._setup_from_response(res)
        return True

    @affirm(is_not_deleted, is_created)
    def delete(self) -> bool:
        """
        Deletes the run.
        """
        delete_run(self._client, run_id=self._id, cluster_id=self.cluster_id)
        self._is_deleted = True
        return True

    def __repr__(self) -> str:
        """Pretty printed output detailing current state of the run.

        Primarily for use as a quick way to execute objects in juyterlab and
        output some info about the run.

        Returns
        -------
        str
            A string detailing the current state of the run.
        """

        # TODO: match values w/ Session object for consistency:
        # - desired_status v desired_state
        # - self.resources.instance_type v self.instance_type
        if self.exists:
            res = textwrap.dedent(
                f"""\
                {self.__class__.__name__}(
                    {"name": <15} = \"{self.name}\",
                    {"description": <15} = \"{self.description}\",
                    {"run_command": <15} = \"{self.run_command}\",
                    {"num_experiments": <15} = \"{len(self.experiments)}\",
                    {"status": <15} = \"{self.status.value}\",
                    {"desired_status": <15} = \"{self.desired_status.value}",
                    {"user": <15} = \"{self.user.user_id}\",
                    {"created_at": <15} = \"{self.created_at}\",
                    {"cluster": <15} = \"{self.cluster_id}\",
                    {"datastore": <15} = \"{self.datastore}",
                )"""
            )
        else:
            res = textwrap.dedent(
                f"""\
                {self.__class__.__name__}(
                    {"name": <15} = \"{self.name}\",
                    {"description": <15} = \"{self.description}\",
                    {"run_command": <15} = \"{self.run_command}\",
                    {"num_experiments": <15} = \"{len(self.experiments)}\",
                    {"status": <15} = \"{self.status.value}\",
                    {"user": <15} = \"{self.user.user_id}\",
                    {"cluster": <15} = \"{self.cluster_id}\",
                    {"datastore": <15} = \"{self.datastore}",
                )"""
            )
        return res

    def __contains__(self, item: str) -> bool:
        """Check if an experiment with the provided name exists in the Run.

        Parameters
        ----------
        item
            The name of the experiment to check for.

        Returns
        -------
        bool
            True if the experiment exists, otherwise False.
        """
        return item in [exp.name for exp in self.experiments]

    def __len__(self) -> int:
        """Check how many experiments are recorded in the Run.

        Returns
        -------
        int
            Number of experiments recorded in the Run.
        """
        return len(self.experiments)

    def __iter__(self) -> Iterable['Experiment']:
        """Allows for iteration over the run, yielding every Experiment object recorded.
        """
        for exp in self.experiments:
            yield exp

    def _unshallow(self):
        """ If the object is a shallow (i.e. only has an id and `_is_shallow` attribute is True)
        object, this method can be triggered to get the full object from the BE. It is designed
        to be called only from the `is_not_shallow` decorator and should not be called directly.
        """
        if not self._is_shallow:
            raise RuntimeError('Run is already unshallow')
        if not hasattr(self, '_id') or self._id is None:
            raise RuntimeError("Cannot unshallow resource without a valid run id")
        self._is_shallow = False
        run = get_run_from_id(self._client, run_id=self._id, cluster_id=self.cluster_id)
        self._setup_from_response(run)

    @affirm(is_not_deleted, is_created)
    def cancel(self) -> bool:
        """Requests the grid platform cancels every Experiment in the Run.

        Returns
        -------
        List[str]
            a collection of experiment names if each experiment was successfully
            cancelled or had previously reached a terminal status ("CANCELLED",
            "FAILED", "COMPLETED").
        """
        if self._created_at is None:
            raise RuntimeError(f"cannot a run which has not been started.")
        updated_run = cancel_run(client=self._client, cluster_id=self.cluster_id, run_id=self._id)
        self._setup_from_response(updated_run)
        return True

    @property
    @affirm(is_not_deleted, is_created)
    def artifacts(self) -> Iterator['Artifact']:
        """Request a list of artifacts created by all experiments in the run.

        Yields
        -------
        Iterator[Artifact]
            Each artifact contains its filename and URL for downloading.
        """
        for experiment in self.experiments:
            for artifact in experiment.artifacts:
                yield artifact


def list_runs(cluster_id: Optional[str] = None, is_global: bool = False, query: str = None) -> List[Run]:
    """
    List all runs on the cluster with the provided cluster id

    Parameters
    ----------
    cluster_id:
        The id of the cluster to list runs for. If None, the default cluster (as listed
        in settings.json) is used.
    is_global:
        If True, return runs of all users in the team
	name:
		if specified, will filter by run name startswith
    """
    cluster_id = cluster_id or env.CONTEXT
    client = GridRestClient(api_client=create_swagger_client())
    users = {}
    if is_global:
        from grid.cli.core import Team
        # TODO - this needs to be moved to REST
        for team in Team.get_all():
            for user_details in team.data.get("members", {}):
                users[user_details["id"]] = user_details
    all_runs = []
    for run_resp in rest_list_runs(client=client, cluster_id=cluster_id, user_ids=list(users.keys()), query=query):
        run_obj = Run(name=SPECIAL_NAME_TO_SKIP_OBJECT_INIT)
        fill_object_from_response(run_obj, run_resp)
        user_details = users.get(run_obj.user.user_id, {})
        user = User(
            user_id=user_details.get("id"),
            username=user_details.get("username"),
            first_name=user_details.get("firstName"),
            last_name=user_details.get("lastName")
        )
        run_obj._user = user
        all_runs.append(run_obj)
    return all_runs


class RunState(Enum):
    """
    Enumeration of the possible run statuses.
    """
    UNSPECIFIED = "unspecified"
    BUILDING = "building"
    PENDING = "pending"
    RUNNING = "running"
    SUCCEEDED = "succeeded"
    FAILED = "failed"
    CANCELLED = "cancelled"
    DELETED = "deleted"

    @classmethod
    def from_api_spec(cls, status: V1RunState) -> 'RunState':
        parsed = str(status).lower().split('_', maxsplit=2)[-1]
        if parsed == 'canceled':  # TODO - spelling mistake in BE
            return cls.CANCELLED
        return cls(parsed)



from datetime import datetime
from typing import Dict, List, Optional

from grid.cli.core.artifact import Artifact
from grid.cli.core.base import GridObject
from grid.cli.exceptions import ResourceNotFound


class Experiment(GridObject):
    """
    Grid experiment object. This object contains all the properties that a
    given artifact should have. This also encapsulates special methods
    for interactive with experiment properties.

    Parameters
    ----------
    name: str
        Experiment name
    """
    terminal_statuses = ("failed", "succeeded", "cancelled")

    def __init__(self, name: Optional[str] = None, identifier: Optional[str] = None):
        self._data: Optional[Dict[str, str]] = {}
        if identifier:
            self.identifier = identifier
        elif not name:
            raise RuntimeError("Either name or identifier is required")
        self.name = name
        super().__init__()

        # get experiment ID upon instantiation if class doesn't have one
        if not hasattr(self, "identifier"):
            self.identifier = self._experiment_id_from_name(self.name)

        # log handling attributes
        self._archive_log_current_page: int
        self._archive_log_total_pages: int

    def _experiment_id_from_name(self, name: str) -> str:
        """Retrieves experiment ID from an experiment name."""

        # User can pass experiments as username:experiment_name to fetch other users experiments
        username = None
        split = name.split(":")
        if len(split) > 2:
            raise ValueError(f"Error while parsing {name}. Use the format <username>:<experiment-name>")
        elif len(split) == 2:
            username = split[0]
            name = split[1]

        query = """
        query ($experimentName: String!, $username: String) {
            getExperimentId(experimentName: $experimentName, username: $username) {
                success
                message
                experimentId
            }
        }
        """
        params = {"experimentName": name, "username": username}
        result = self.client.query(query, **params)

        if not result["getExperimentId"]["success"]:
            if "Cannot find experiment" in result["getExperimentId"]["message"]:
                raise ResourceNotFound(
                    f"Experiment {name} does not exist\nIf you wish to fetch an experiment for somebody in your team, use <username>:<experiment-name> format"
                )
            raise ValueError(f"{result['getExperimentId']['message']}")

        return result["getExperimentId"]["experimentId"]

    def refresh(self) -> None:
        """
        Updates object metadata. This makes a query to Grid to fetch the
        object"s latest data.
        """
        query = """
        query GetExperimentDetails ($experimentId: ID!) {
            getExperimentDetails(experimentId: $experimentId) {
                name
                githubId
                desiredState
                commitSha
                entrypoint
                invocationCommands
                createdAt
                startedRunningAt
                finishedAt
                run {
                    runId
                    name
                }
                parameters {
                    name
                    value
                }
            }
        }
        """
        result = self.client.query(query, experimentId=self.identifier)
        self._data = result["getExperimentDetails"]
        self._update_meta()

    def _archive_logs(self,
                      start: Optional[datetime] = None,
                      stop: Optional[datetime] = None,
                      limit: int = 100) -> List[Dict[str, str]]:
        """
        Retrieves logs from the archives. Logs from the archives can only
        be retrieved using a time window.

        Parameters
        ----------
        start: datetime
            Start period

        stop: datetime
            Stop period

        limit: int, default 100
            Number of lines to return. This only filters lines in memory.

        Returns
        -------
        List[Dict[str, str]]
            List of log records
        """
        query = """
        query ($experimentId: ID!, $start: DateTime, $stop: DateTime) {
            getExperimentLogs(experimentId: $experimentId, start: $start, stop: $stop) {
                success
                message

                lines {
                    message
                    timestamp
                }
            }
        }
        """
        params = {"experimentId": self.identifier, "start": start, "stop": stop}
        result = self.client.query(query, **params)

        data = result["getExperimentLogs"]

        # check if request was successful
        if not data["success"]:
            raise ValueError(f"Failed to fetch logs. Error: {data['message']}")

        # set metadata for pagination
        if not data['lines']:
            return []
        self._archive_logs_latest_timestamp = max({l["timestamp"] for l in data["lines"]})

        # return just log lines; filter based on limit
        # TODO: this operation is better placed in the backend. This will reduce
        # data transfer and be faster overall.
        output = data["lines"]
        if limit is not None and limit > 0:
            output = output[-limit:]

        # add type identifier
        for log in output:
            log["type"] = "experiment"

        return output

    def _live_logs(self, limit: int) -> Dict[str, str]:
        """
        Streams real-time experiment logs.

        Yields
        ------
        Dict[str, str]
            Log lines
        """
        subscription = """
        subscription ($experimentId: ID!, $limit:Int) {
            getLiveExperimentLogs(
                experimentId: $experimentId,
                limit: $limit) {
                    message
                    timestamp
            }
        }
        """
        params = {"experimentId": self.identifier, "limit": limit}
        stream = self.client.subscribe(query=subscription, **params)
        for element in stream:
            for entry in element["getLiveExperimentLogs"]:
                yield {"type": "experiment", **entry}

    def _live_build_logs(self, limit: int) -> Dict[str, str]:
        """
        Streams build logs from subscription API.

        Yields
        ------
        Dict[str, str]
            Log lines
        """
        subscription = """
        subscription ($experimentId: ID!, $tailLines:Int) {
            getLiveBuildLogs(
                experimentId: $experimentId,
                tailLines: $tailLines) {
                    message
                    timestamp
            }
        }
        """
        params = {"experimentId": self.identifier, "tailLines": limit}
        stream = self.client.subscribe(query=subscription, **params)
        for element in stream:
            yield {"type": "build", **element["getLiveBuildLogs"]}

    def _archive_build_logs(self,
                            start: Optional[datetime] = None,
                            stop: Optional[datetime] = None,
                            limit: int = 100) -> List[Dict[str, str]]:
        """
        Retrievies build logs from the archive.

        Yields
        ------
        Dict[str, str]
            Log lines
        """
        subscription = """
        query ($experimentId: ID!, $start: DateTime, $stop: DateTime) {
            getBuildLogs(
                experimentId: $experimentId,
                start: $start,
                stop: $stop) {
                    message
                    timestamp
            }
        }
        """
        params = {"experimentId": self.identifier, "start": start, "stop": stop}
        result = self.client.query(query=subscription, **params)

        data = result["getBuildLogs"]

        # add type identifier
        for log in data:
            log["type"] = "build"
        if limit is not None and limit > 0:
            data = data[-limit:]
        return data

    def logs(self, build_logs: Optional[bool] = None, tail_lines: Optional[int] = None) -> Dict[str, str]:
        """
        Experiment logs generator; combines both stdout and stderr logs.

        Parameters
        ----------
        build_logs: bool, default None
            If result should include build logs. When left to `None`, this will automatically
            include build logs in output when `Experiment.is_queued` evaluates to `True`.
            If `build_logs=True` is passed, then this includes build logs in the output
            by default.
        tail_lines: int, default None
            Number of lines to fetch from end. If subscribing for logs on a running
            experiment, this will fetch up to this number of lines.

        Yields
        ------
        Dict[str, str]
            Log entry
        """
        if build_logs is not False:
            # TODO (luiscape): pagination summary API is required for pagination
            if self.is_queued:
                for entry in self._live_build_logs(limit=tail_lines):
                    yield entry
            else:
                for entry in self._archive_build_logs(limit=tail_lines):
                    yield entry
        if self.is_terminal:
            # adds limit for a better user experience; this is done to
            # prevent overwheming clients with all logs all the time
            if tail_lines is None:
                tail_lines = 100

            for entry in self._archive_logs(limit=tail_lines):
                yield entry
        elif self.is_running:
            for entry in self._live_logs(limit=tail_lines):
                yield entry

    @property
    def is_terminal(self):
        """Determines if an experiment is in terminal status."""
        return self.status in self.terminal_statuses

    @property
    def is_running(self):
        """Determines if an experiment is in a running status."""
        return self.status == "running"

    @property
    def is_queued(self):
        """Determines if an experiment is in a queued status."""
        return self.status == "queued"

    @property
    def is_preparing(self):
        """Represents states prior to an experiment running"""
        return self.status in ("queued", "pending")

    @property
    def status(self) -> str:
        query = """
        query ($experimentId: ID!) {
            getExperimentDetails(experimentId: $experimentId) {
                status
            }
        }
        """
        result = self.client.query(query, experimentId=self.identifier)
        return result["getExperimentDetails"]["status"]

    @property
    def artifacts(self) -> List[Artifact]:
        """Fetches artifacts from a given experiments. Artifacts are"""
        query = """
        query ($experimentId: ID!) {
            getArtifacts(experimentId: $experimentId) {
                signedUrl
                downloadToPath
                downloadToFilename
            }
        }
        """
        result = self.client.query(query, experimentId=self.identifier)
        return [Artifact(*a.values()) for a in result.get("getArtifacts")]



from grid.sdk.api import (
    get_instance_types,
    list_clusters,
)
from grid.sdk.runs import Run, Actions, Resources, ScratchSpace, list_runs
from grid.sdk.datastores import Datastore, list_datastores
from grid.sdk.sessions import Session, list_sessions
from grid.sdk.experiments import Experiment
from grid.sdk.login import login
from grid.sdk.artifacts import Artifact, download_artifacts, list_artifacts

__all__ = [
    "login", "Datastore", "list_datastores", "Run", "list_runs", "Actions", "ScratchSpace", "Resources", "Experiment",
    "get_instance_types", "list_clusters", "list_sessions", "Session", "Artifact", "download_artifacts",
    "list_artifacts"
]



from grid.openapi import (
    ApiClient, AuthServiceApi, ClusterServiceApi, DatastoreServiceApi, RunServiceApi, SAMLOrganizationsServiceApi,
    SessionServiceApi, TensorboardServiceApi, ExperimentServiceApi, CredentialServiceApi
)


class GridRestClient(
    AuthServiceApi, ClusterServiceApi, DatastoreServiceApi, RunServiceApi, SAMLOrganizationsServiceApi,
    SessionServiceApi, TensorboardServiceApi, ExperimentServiceApi, CredentialServiceApi
):
    api_client: ApiClient

    def __init__(self, api_client: ApiClient):  # skipcq: PYL-W0231
        self.api_client = api_client



from grid.sdk.datastores import DATASTORE_FSX_THROUGHPUT_ALIASES
from grid.cli.observables.base import create_table
from grid.sdk.datastores import Datastore


def datastore_uploaded_info_table(dstore: Datastore):

    table_dict = {
        'Name': dstore.name,
        'Cluster ID': dstore.cluster_id,
        'Version': str(dstore.version),
        'Size': str(dstore.size),
        'Created At': f'{dstore.created_at:%Y-%m-%d %H:%M}'
    }
    if dstore._fsx_enabled:
        fsx_throughput_mbs_tib = dstore._fsx_throughput_mbs_tib
        fsx_capacity = dstore._fsx_capacity_gib
        fsx_throughput_alias = {v: k
                                for k, v in DATASTORE_FSX_THROUGHPUT_ALIASES.items()}.get(fsx_throughput_mbs_tib, "")

        table_dict.update({
            'HPD Throughput': f'{fsx_throughput_alias} ({fsx_throughput_mbs_tib}MB/s/TiB)',
            'HPD Capacity': f'{fsx_capacity}GiB',
            'Total Throughput': f'{fsx_throughput_mbs_tib * fsx_capacity / 1000}MB/s'
        })

    table = create_table([key for key, value in table_dict.items() if value is not None])
    table.add_row(*[value for value in table_dict.values() if value is not None])

    return table



from collections import defaultdict

import click
from rich.console import Console

from grid import get_instance_types
from grid.cli import rich_click
from grid.cli.observables import BaseObservable
from grid.sdk import env
from grid.sdk.rest.exceptions import GridException
from grid.cli.cli.grid_run import _aws_node_to_nickname


@rich_click.command('instance-types')
@click.option(
    '--cluster',
    'cluster_id',
    type=str,
    required=False,
    default=env.CONTEXT,
    help='Cluster ID whence the instance types needs to be fetched. (Bring Your Own Cloud Customers Only).'
)
def instance_types(cluster_id: str):
    """List the compute node instance types which are available for computation.

    For bring your own cloud customers, the instance types available are
    defined by the organizational administrators who created the cluster.
    """
    try:
        instances = get_instance_types(cluster_id=cluster_id)
    except GridException as e:
        raise click.ClickException(str(e))
    except Exception as e:
        raise click.ClickException(f"Could not fetch instance types: {e}")

    instance_dict = defaultdict(dict)
    table = BaseObservable.create_table(
        columns=['Instance Type', 'On-demand Cost', 'Spot Cost', 'CPU', 'GPU', 'Memory', 'Accelerator']
    )

    for i in instances:
        if i.spot:
            instance_dict[i.name]['spot_cost'] = i.hourly_cost
        else:
            instance_dict[i.name]['on_demand_cost'] = i.hourly_cost
        # these fields shouldn't be different for spot/on-demand
        instance_dict[i.name]['cpu'] = i.cpu
        instance_dict[i.name]['gpu'] = i.gpu
        instance_dict[i.name]['memory'] = i.memory

    for name, det in instance_dict.items():
        on_demand_cost = str(det.get('on_demand_cost') or "Not Available")
        spot_cost = str(det.get('spot_cost') or "Not Available")
        # We'll extract accelerator type from nickname
        nickname = _aws_node_to_nickname().get(name)
        accelerator = "" if nickname is None else nickname.split('_')[1]
        table.add_row(name, on_demand_cost, spot_cost, det['cpu'], det['gpu'], det['memory'], accelerator)
    console = Console()
    console.print(table)



import abc

import click
from rich.console import Console
from rich.table import Table


class Formatable(abc.ABC):
    @abc.abstractmethod
    def as_table(self) -> Table:
        pass

    @abc.abstractmethod
    def as_json(self) -> str:
        pass


def print_to_console(ctx: click.Context, formatable: Formatable):
    obj = ctx.obj or {}
    if 'console' not in obj:
        console = Console()
        obj['console'] = console
    console: Console = obj['console']
    output_format = obj.get('output', 'console')
    if output_format == "json":
        console.print(formatable.as_json(), highlight=False)
    elif output_format == "console":
        console.print(formatable.as_table())
    else:
        raise ValueError(f"unknown output format {output_format}")
